@@ -1,9 +1,9 @@
 # -*- coding: utf-8 -*-
 #
 # This file is part of Radicale Server - Calendar Server
-# Copyright © 2008-2012 Guillaume Ayoub
 # Copyright © 2008 Nicolas Kandel
 # Copyright © 2008 Pascal Halter
+# Copyright © 2008-2013 Guillaume Ayoub
 #
 # This library is free software: you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -29,27 +29,35 @@ should have been included in this package.
 """
 
 import os
+import sys
 import pprint
 import base64
 import posixpath
 import socket
 import ssl
 import wsgiref.simple_server
+import re
 # Manage Python2/3 different modules
 # pylint: disable=F0401,E0611
 try:
     from http import client
-    from urllib.parse import quote, unquote, urlparse
+    from urllib.parse import unquote, urlparse
 except ImportError:
     import httplib as client
-    from urllib import quote, unquote
+    from urllib import unquote
     from urlparse import urlparse
 # pylint: enable=F0401,E0611
 
-from radicale import acl, config, ical, log, storage, xmlutils
+from . import auth, config, ical, log, rights, storage, xmlutils
 
 
-VERSION = "0.7"
+VERSION = "0.10"
+
+# Standard "not allowed" response that is returned when an authenticated user
+# tries to access information they don't have rights to
+NOT_ALLOWED = (client.FORBIDDEN, {}, None)
+
+WELL_KNOWN_RE = re.compile(r"/.well-known/(carddav|caldav)/?")
 
 
 class HTTPServer(wsgiref.simple_server.WSGIServer, object):
@@ -85,16 +93,21 @@ class HTTPSServer(HTTPServer):
             try:
                 open(filename, "r").close()
             except IOError as exception:
-                log.LOGGER.warn(
+                log.LOGGER.warning(
                     "Error while reading SSL %s %r: %s" % (
                         name, filename, exception))
 
-        self.socket = ssl.wrap_socket(
-            self.socket,
+        ssl_kwargs = dict(
             server_side=True,
             certfile=config.get("server", "certificate"),
             keyfile=config.get("server", "key"),
-            ssl_version=ssl.PROTOCOL_SSLv23)
+            ssl_version=getattr(
+                ssl, config.get("server", "protocol"), ssl.PROTOCOL_SSLv23))
+        # add ciphers argument only if supported (Python 2.7+)
+        if sys.version_info >= (2, 7):
+            ssl_kwargs["ciphers"] = config.get("server", "ciphers") or None
+
+        self.socket = ssl.wrap_socket(self.socket, **ssl_kwargs)
 
         self.server_bind()
         self.server_activate()
@@ -119,8 +132,9 @@ class Application(object):
     def __init__(self):
         """Initialize application."""
         super(Application, self).__init__()
-        self.acl = acl.load()
+        auth.load()
         storage.load()
+        rights.load()
         self.encoding = config.get("encoding", "request")
         if config.getboolean("logging", "full_environment"):
             self.headers_log = lambda environ: environ
@@ -169,16 +183,135 @@ class Application(object):
         trailing_slash = "" if uri == "/" else trailing_slash
         return uri + trailing_slash
 
+    def collect_allowed_items(self, items, user):
+        """Get items from request that user is allowed to access."""
+        read_last_collection_allowed = None
+        write_last_collection_allowed = None
+        read_allowed_items = []
+        write_allowed_items = []
+
+        for item in items:
+            if isinstance(item, ical.Collection):
+                if rights.authorized(user, item, "r"):
+                    log.LOGGER.debug(
+                        "%s has read access to collection %s" %
+                        (user or "Anonymous", item.url or "/"))
+                    read_last_collection_allowed = True
+                    read_allowed_items.append(item)
+                else:
+                    log.LOGGER.debug(
+                        "%s has NO read access to collection %s" %
+                        (user or "Anonymous", item.url or "/"))
+                    read_last_collection_allowed = False
+
+                if rights.authorized(user, item, "w"):
+                    log.LOGGER.debug(
+                        "%s has write access to collection %s" %
+                        (user or "Anonymous", item.url or "/"))
+                    write_last_collection_allowed = True
+                    write_allowed_items.append(item)
+                else:
+                    log.LOGGER.debug(
+                        "%s has NO write access to collection %s" %
+                        (user or "Anonymous", item.url or "/"))
+                    write_last_collection_allowed = False
+            else:
+                # item is not a collection, it's the child of the last
+                # collection we've met in the loop. Only add this item
+                # if this last collection was allowed.
+                if read_last_collection_allowed:
+                    log.LOGGER.debug(
+                        "%s has read access to item %s" %
+                        (user or "Anonymous", item.name))
+                    read_allowed_items.append(item)
+                else:
+                    log.LOGGER.debug(
+                        "%s has NO read access to item %s" %
+                        (user or "Anonymous", item.name))
+
+                if write_last_collection_allowed:
+                    log.LOGGER.debug(
+                        "%s has write access to item %s" %
+                        (user or "Anonymous", item.name))
+                    write_allowed_items.append(item)
+                else:
+                    log.LOGGER.debug(
+                        "%s has NO write access to item %s" %
+                        (user or "Anonymous", item.name))
+
+        return read_allowed_items, write_allowed_items
+
     def __call__(self, environ, start_response):
         """Manage a request."""
         log.LOGGER.info("%s request at %s received" % (
-                environ["REQUEST_METHOD"], environ["PATH_INFO"]))
+            environ["REQUEST_METHOD"], environ["PATH_INFO"]))
         headers = pprint.pformat(self.headers_log(environ))
         log.LOGGER.debug("Request headers:\n%s" % headers)
 
-        # Sanitize request URI
-        environ["PATH_INFO"] = self.sanitize_uri(environ["PATH_INFO"])
-        log.LOGGER.debug("Sanitized path: %s", environ["PATH_INFO"])
+        base_prefix = config.get("server", "base_prefix")
+        if environ["PATH_INFO"].startswith(base_prefix):
+            # Sanitize request URI
+            environ["PATH_INFO"] = self.sanitize_uri(
+                "/%s" % environ["PATH_INFO"][len(base_prefix):])
+            log.LOGGER.debug("Sanitized path: %s", environ["PATH_INFO"])
+        elif config.get("server", "can_skip_base_prefix"):
+            log.LOGGER.debug(
+                "Skipped already sanitized path: %s", environ["PATH_INFO"])
+        else:
+            # Request path not starting with base_prefix, not allowed
+            log.LOGGER.debug(
+                "Path not starting with prefix: %s", environ["PATH_INFO"])
+            environ["PATH_INFO"] = None
+
+        path = environ["PATH_INFO"]
+
+        # Get function corresponding to method
+        function = getattr(self, environ["REQUEST_METHOD"].lower())
+
+        # Ask authentication backend to check rights
+        authorization = environ.get("HTTP_AUTHORIZATION", None)
+
+        if authorization:
+            authorization = authorization.lstrip("Basic").strip()
+            user, password = self.decode(base64.b64decode(
+                authorization.encode("ascii")), environ).split(":", 1)
+        else:
+            user = environ.get("REMOTE_USER")
+            password = None
+
+        well_known = WELL_KNOWN_RE.match(path)
+        if well_known:
+            redirect = config.get("well-known", well_known.group(1))
+            try:
+                redirect = redirect % ({"user": user} if user else {})
+            except KeyError:
+                status = client.UNAUTHORIZED
+                headers = {
+                    "WWW-Authenticate":
+                    "Basic realm=\"%s\"" % config.get("server", "realm")}
+                log.LOGGER.info(
+                    "Refused /.well-known/ redirection to anonymous user")
+            else:
+                status = client.SEE_OTHER
+                log.LOGGER.info("/.well-known/ redirection to: %s" % redirect)
+                if sys.version_info < (3, 0):
+                    redirect = redirect.encode(self.encoding)
+                headers = {"Location": redirect}
+            status = "%i %s" % (
+                status, client.responses.get(status, "Unknown"))
+            start_response(status, list(headers.items()))
+            return []
+
+        is_authenticated = auth.is_authenticated(user, password)
+        is_valid_user = is_authenticated or not user
+
+        if is_valid_user:
+            items = ical.Collection.from_path(
+                path, environ.get("HTTP_DEPTH", "0"))
+            read_allowed_items, write_allowed_items = \
+                self.collect_allowed_items(items, user)
+        else:
+            read_allowed_items, write_allowed_items = None, None
 
         # Get content
         content_length = int(environ.get("CONTENT_LENGTH") or 0)
@@ -189,72 +322,26 @@ class Application(object):
         else:
             content = None
 
-        # Find collection(s)
-        items = ical.Collection.from_path(
-            environ["PATH_INFO"], environ.get("HTTP_DEPTH", "0"))
-
-        # Get function corresponding to method
-        function = getattr(self, environ["REQUEST_METHOD"].lower())
-
-        # Check rights
-        if not items or not self.acl or function == self.options:
-            # No collection, or no acl, or OPTIONS request: don't check rights
-            status, headers, answer = function(environ, items, content, None)
+        if is_valid_user and (
+                (read_allowed_items or write_allowed_items) or
+                (is_authenticated and function == self.propfind) or
+                function == self.options):
+            status, headers, answer = function(
+                environ, read_allowed_items, write_allowed_items, content,
+                user)
         else:
-            # Ask authentication backend to check rights
-            authorization = environ.get("HTTP_AUTHORIZATION", None)
+            status, headers, answer = NOT_ALLOWED
 
-            if authorization:
-                auth = authorization.lstrip("Basic").strip().encode("ascii")
-                user, password = self.decode(
-                    base64.b64decode(auth), environ).split(":")
-            else:
-                user = password = None
-
-            last_allowed = None
-            collections = []
-            for collection in items:
-                if not isinstance(collection, ical.Collection):
-                    if last_allowed:
-                        collections.append(collection)
-                    continue
-
-                if collection.owner in acl.PUBLIC_USERS:
-                    log.LOGGER.info("Public collection")
-                    collections.append(collection)
-                    last_allowed = True
-                else:
-                    log.LOGGER.info(
-                        "Checking rights for collection owned by %s" % (
-                            collection.owner or "nobody"))
-                    if self.acl.has_right(collection.owner, user, password):
-                        log.LOGGER.info(
-                            "%s allowed" % (user or "Anonymous user"))
-                        collections.append(collection)
-                        last_allowed = True
-                    else:
-                        log.LOGGER.info(
-                            "%s refused" % (user or "Anonymous user"))
-                        last_allowed = False
-
-            if collections:
-                # Collections found
-                status, headers, answer = function(
-                    environ, collections, content, user)
-            elif user and last_allowed is None:
-                # Good user and no collections found, redirect user to home
-                location = "/%s/" % str(quote(user))
-                log.LOGGER.info("redirecting to %s" % location)
-                status = client.FOUND
-                headers = {"Location": location}
-                answer = "Redirecting to %s" % location
-            else:
-                # Unknown or unauthorized user
-                status = client.UNAUTHORIZED
-                headers = {
-                    "WWW-Authenticate":
-                    "Basic realm=\"Radicale Server - Password Required\""}
-                answer = None
+        if ((status, headers, answer) == NOT_ALLOWED and
+                not auth.is_authenticated(user, password) and
+                config.get("auth", "type") != "None"):
+            # Unknown or unauthorized user
+            log.LOGGER.info("%s refused" % (user or "Anonymous user"))
+            status = client.UNAUTHORIZED
+            headers = {
+                "WWW-Authenticate":
+                "Basic realm=\"%s\"" % config.get("server", "realm")}
+            answer = None
 
         # Set content length
         if answer:
@@ -262,6 +349,10 @@ class Application(object):
                 "Response content:\n%s" % self.decode(answer, environ))
             headers["Content-Length"] = str(len(answer))
 
+        if config.has_section("headers"):
+            for key in config.options("headers"):
+                headers[key] = config.get("headers", key)
+
         # Start response
         status = "%i %s" % (status, client.responses.get(status, "Unknown"))
         log.LOGGER.debug("Answer status: %s" % status)
@@ -273,9 +364,13 @@ class Application(object):
     # All these functions must have the same parameters, some are useless
     # pylint: disable=W0612,W0613,R0201
 
-    def delete(self, environ, collections, content, user):
+    def delete(self, environ, read_collections, write_collections, content,
+               user):
         """Manage DELETE request."""
-        collection = collections[0]
+        if not len(write_collections):
+            return NOT_ALLOWED
+
+        collection = write_collections[0]
 
         if collection.path == environ["PATH_INFO"].strip("/"):
             # Path matching the collection, the collection must be deleted
@@ -285,19 +380,18 @@ class Application(object):
             item = collection.get_item(
                 xmlutils.name_from_path(environ["PATH_INFO"], collection))
 
-        # Evolution bug workaround
-        etag = environ.get("HTTP_IF_MATCH", item.etag).replace("\\", "")
-        if item and etag == item.etag:
-            # No ETag precondition or precondition verified, delete item
-            answer = xmlutils.delete(environ["PATH_INFO"], collection)
-            status = client.NO_CONTENT
-        else:
-            # No item or ETag precondition not verified, do not delete item
-            answer = None
-            status = client.PRECONDITION_FAILED
-        return status, {}, answer
+        if item:
+            # Evolution bug workaround
+            etag = environ.get("HTTP_IF_MATCH", item.etag).replace("\\", "")
+            if etag == item.etag:
+                # No ETag precondition or precondition verified, delete item
+                answer = xmlutils.delete(environ["PATH_INFO"], collection)
+                return client.OK, {}, answer
 
-    def get(self, environ, collections, content, user):
+        # No item or ETag precondition not verified, do not delete item
+        return client.PRECONDITION_FAILED, {}, None
+
+    def get(self, environ, read_collections, write_collections, content, user):
         """Manage GET request.
 
         In Radicale, GET requests create collections when the URL is not
@@ -308,10 +402,14 @@ class Application(object):
         # Display a "Radicale works!" message if the root URL is requested
         if environ["PATH_INFO"] == "/":
             headers = {"Content-type": "text/html"}
-            answer = "<!DOCTYPE html>\n<title>Radicale</title>Radicale works!"
+            answer = b"<!DOCTYPE html>\n<title>Radicale</title>Radicale works!"
             return client.OK, headers, answer
 
-        collection = collections[0]
+        if not len(read_collections):
+            return NOT_ALLOWED
+
+        collection = read_collections[0]
+
         item_name = xmlutils.name_from_path(environ["PATH_INFO"], collection)
 
         if item_name:
@@ -328,7 +426,15 @@ class Application(object):
         else:
             # Create the collection if it does not exist
             if not collection.exists:
-                collection.write()
+                if collection in write_collections:
+                    log.LOGGER.debug(
+                        "Creating collection %s" % collection.name)
+                    collection.write()
+                else:
+                    log.LOGGER.debug(
+                        "Collection %s not available and could not be created "
+                        "due to missing write rights" % collection.name)
+                    return NOT_ALLOWED
 
             # Get whole collection
             answer_text = collection.text
@@ -341,14 +447,21 @@ class Application(object):
         answer = answer_text.encode(self.encoding)
         return client.OK, headers, answer
 
-    def head(self, environ, collections, content, user):
+    def head(self, environ, read_collections, write_collections, content,
+             user):
         """Manage HEAD request."""
-        status, headers, answer = self.get(environ, collections, content, user)
+        status, headers, answer = self.get(
+            environ, read_collections, write_collections, content, user)
         return status, headers, None
 
-    def mkcalendar(self, environ, collections, content, user):
+    def mkcalendar(self, environ, read_collections, write_collections, content,
+                   user):
         """Manage MKCALENDAR request."""
-        collection = collections[0]
+        if not len(write_collections):
+            return NOT_ALLOWED
+
+        collection = write_collections[0]
+
         props = xmlutils.props_from_request(content)
         timezone = props.get("C:calendar-timezone")
         if timezone:
@@ -357,12 +470,17 @@ class Application(object):
         with collection.props as collection_props:
             for key, value in props.items():
                 collection_props[key] = value
-        collection.write()
+            collection.write()
         return client.CREATED, {}, None
 
-    def mkcol(self, environ, collections, content, user):
+    def mkcol(self, environ, read_collections, write_collections, content,
+              user):
         """Manage MKCOL request."""
-        collection = collections[0]
+        if not len(write_collections):
+            return NOT_ALLOWED
+
+        collection = write_collections[0]
+
         props = xmlutils.props_from_request(content)
         with collection.props as collection_props:
             for key, value in props.items():
@@ -370,9 +488,14 @@ class Application(object):
         collection.write()
         return client.CREATED, {}, None
 
-    def move(self, environ, collections, content, user):
+    def move(self, environ, read_collections, write_collections, content,
+             user):
         """Manage MOVE request."""
-        from_collection = collections[0]
+        if not len(write_collections):
+            return NOT_ALLOWED
+
+        from_collection = write_collections[0]
+
         from_name = xmlutils.name_from_path(
             environ["PATH_INFO"], from_collection)
         if from_name:
@@ -385,9 +508,12 @@ class Application(object):
                     to_path, to_name = to_url.rstrip("/").rsplit("/", 1)
                     to_collection = ical.Collection.from_path(
                         to_path, depth="0")[0]
-                    to_collection.append(to_name, item.text)
-                    from_collection.remove(from_name)
-                    return client.CREATED, {}, None
+                    if to_collection in write_collections:
+                        to_collection.append(to_name, item.text)
+                        from_collection.remove(from_name)
+                        return client.CREATED, {}, None
+                    else:
+                        return NOT_ALLOWED
                 else:
                     # Remote destination server, not supported
                     return client.BAD_GATEWAY, {}, None
@@ -398,35 +524,49 @@ class Application(object):
             # Moving collections, not supported
             return client.FORBIDDEN, {}, None
 
-    def options(self, environ, collections, content, user):
+    def options(self, environ, read_collections, write_collections, content,
+                user):
         """Manage OPTIONS request."""
         headers = {
-            "Allow": "DELETE, HEAD, GET, MKCALENDAR, MKCOL, MOVE, " \
-                "OPTIONS, PROPFIND, PROPPATCH, PUT, REPORT",
+            "Allow": ("DELETE, HEAD, GET, MKCALENDAR, MKCOL, MOVE, "
+                      "OPTIONS, PROPFIND, PROPPATCH, PUT, REPORT"),
             "DAV": "1, 2, 3, calendar-access, addressbook, extended-mkcol"}
         return client.OK, headers, None
 
-    def propfind(self, environ, collections, content, user):
+    def propfind(self, environ, read_collections, write_collections, content,
+                 user):
         """Manage PROPFIND request."""
+        # Rights is handled by collection in xmlutils.propfind
         headers = {
             "DAV": "1, 2, 3, calendar-access, addressbook, extended-mkcol",
             "Content-Type": "text/xml"}
+        collections = set(read_collections + write_collections)
         answer = xmlutils.propfind(
             environ["PATH_INFO"], content, collections, user)
         return client.MULTI_STATUS, headers, answer
 
-    def proppatch(self, environ, collections, content, user):
+    def proppatch(self, environ, read_collections, write_collections, content,
+                  user):
         """Manage PROPPATCH request."""
-        collection = collections[0]
-        answer = xmlutils.proppatch(environ["PATH_INFO"], content, collection)
+        if not len(write_collections):
+            return NOT_ALLOWED
+
+        collection = write_collections[0]
+
+        answer = xmlutils.proppatch(
+            environ["PATH_INFO"], content, collection)
         headers = {
             "DAV": "1, 2, 3, calendar-access, addressbook, extended-mkcol",
             "Content-Type": "text/xml"}
         return client.MULTI_STATUS, headers, answer
 
-    def put(self, environ, collections, content, user):
+    def put(self, environ, read_collections, write_collections, content, user):
         """Manage PUT request."""
-        collection = collections[0]
+        if not len(write_collections):
+            return NOT_ALLOWED
+
+        collection = write_collections[0]
+
         collection.set_mimetype(environ.get("CONTENT_TYPE"))
         headers = {}
         item_name = xmlutils.name_from_path(environ["PATH_INFO"], collection)
@@ -434,24 +574,37 @@ class Application(object):
 
         # Evolution bug workaround
         etag = environ.get("HTTP_IF_MATCH", "").replace("\\", "")
+        match = environ.get("HTTP_IF_NONE_MATCH", "") == "*"
         if (not item and not etag) or (
-            item and ((etag or item.etag) == item.etag)):
+                item and ((etag or item.etag) == item.etag) and not match):
             # PUT allowed in 3 cases
             # Case 1: No item and no ETag precondition: Add new item
             # Case 2: Item and ETag precondition verified: Modify item
             # Case 3: Item and no Etag precondition: Force modifying item
             xmlutils.put(environ["PATH_INFO"], content, collection)
             status = client.CREATED
-            headers["ETag"] = collection.get_item(item_name).etag
+            # Try to return the etag in the header.
+            # If the added item doesn't have the same name as the one given
+            # by the client, then there's no obvious way to generate an
+            # etag, we can safely ignore it.
+            new_item = collection.get_item(item_name)
+            if new_item:
+                headers["ETag"] = new_item.etag
         else:
             # PUT rejected in all other cases
             status = client.PRECONDITION_FAILED
         return status, headers, None
 
-    def report(self, environ, collections, content, user):
+    def report(self, environ, read_collections, write_collections, content,
+               user):
         """Manage REPORT request."""
-        collection = collections[0]
+        if not len(read_collections):
+            return NOT_ALLOWED
+
+        collection = read_collections[0]
+
         headers = {"Content-Type": "text/xml"}
+
         answer = xmlutils.report(environ["PATH_INFO"], content, collection)
         return client.MULTI_STATUS, headers, answer
 
